#include <blackfin.h>
#include <builtins.h>
#include <sys/exception.h>
#include <stdint.h>
#include <string.h>
#include "bfin_twi.h"

#define TWI_COUNT(x) (DCNT & ((x) << 6))

typedef struct
{
	int32_t             state;
	BFIN_TWI_XFER_ITEM* p_item;
	int32_t             item_count;
} twi_xfer_param;

static twi_xfer_param sf_twi_xfer_param;
static BFIN_TWI_XFER_ITEM sf_twi_xfer_items[2];

EX_INTERRUPT_HANDLER(s_bfin_twi_isr);
static void s_bfin_twi_isr(void);

void e_bfin_twi_init(void)
{
	memset(&sf_twi_xfer_param, 0, sizeof(sf_twi_xfer_param));
	
	*pTWI_CONTROL = 0;
	ssync();
	*pTWI_CONTROL = 10U;
	*pTWI_CLKDIV  = CLKHI(8) | CLKLOW(17);
	*pTWI_INT_MASK = RCVSERV | XMTSERV | MERR | MCOMP;
	*pTWI_INT_STAT = *pTWI_INT_STAT;
	*pTWI_FIFO_CTL = XMTFLUSH | RCVFLUSH;
	ssync();
	*pTWI_FIFO_CTL = 0;
	
	/* 荞݃nhIVG12ɓo^ */
	register_handler(ik_ivg12, s_bfin_twi_isr);
	*pSIC_IMASK0 |= IRQ_TWI;

	*pTWI_CONTROL |= TWI_ENA;
}

static void s_bfin_twi_master_xfer_start(BFIN_TWI_XFER_ITEM* p_item)
{
	uint8_t* lp_buffer = p_item->p_buffer;
	uint16_t l_count   = p_item->count;
	
	if((BFIN_TWI_MODE_WRITE == p_item->access_mode) || (BFIN_TWI_MODE_WRITE_REPEAT == p_item->access_mode))
	{
		p_item->p_buffer = &p_item->p_buffer[1];
		p_item->count--;
	}	
	
	*pTWI_INT_STAT = MCOMP | MERR | XMTSERV | RCVSERV;
	*pTWI_MASTER_STAT = BUFWRERR | BUFRDERR | LOSTARB | ANAK | DNAK;
	ssync();
	if((BFIN_TWI_MODE_READ == p_item->access_mode) || (BFIN_TWI_MODE_READ_REPEAT == p_item->access_mode))
	{
		// Read
		if(BFIN_TWI_MODE_READ_REPEAT == p_item->access_mode)
		{
			// Repeat
			*pTWI_MASTER_CTL = TWI_COUNT(l_count) | FAST | MDIR | MEN | RSTART;
		}
		else
		{
			// Single
			*pTWI_MASTER_CTL = TWI_COUNT(l_count) | FAST | MDIR | MEN;
		}
	}
	else
	{
		// Write
		*pTWI_XMT_DATA8  = lp_buffer[0];
		if(BFIN_TWI_MODE_WRITE_REPEAT == p_item->access_mode)
		{
			// Repeat
			*pTWI_MASTER_CTL = TWI_COUNT(l_count) | FAST | MEN | RSTART;
		}
		else
		{
			// Single
			*pTWI_MASTER_CTL = TWI_COUNT(l_count) | FAST | MEN;
		}
	}	
}

int32_t e_bfin_twi_master_xfer(uint8_t slave_addr, BFIN_TWI_XFER_ITEM* p_item, int32_t item_count)
{
	int32_t l_lc;
	int32_t l_err = BFIN_TWI_ERR_SUCCESS;
	
	if(item_count <= 0) return BFIN_TWI_ERR_PARAM;

	// Check argument
	for(l_lc = 0; l_lc < item_count; l_lc++)
	{
		if(0u == p_item[l_lc].count)
		{
			l_err = BFIN_TWI_ERR_PARAM;
			break;
		}
		p_item[l_lc].state = 0;
	}
	
	if(BFIN_TWI_ERR_SUCCESS == l_err)
	{
		sf_twi_xfer_param.state      = BFIN_TWI_STATE_XFER;
		sf_twi_xfer_param.p_item     = p_item;
		sf_twi_xfer_param.item_count = item_count - 1;

		*pTWI_MASTER_ADDR = slave_addr;
		s_bfin_twi_master_xfer_start(&p_item[0]);
	}
	else
	{
		sf_twi_xfer_param.state      = BFIN_TWI_STATE_ERROR;
	}
	
	return l_err;
}

int32_t e_bfin_twi_master_write_and_read(uint8_t slave_addr, const uint8_t* p_write_data, uint16_t write_cnt, uint8_t* p_read_data, uint16_t read_cnt)
{
	if((write_cnt == 0) || (read_cnt == 0)) return BFIN_TWI_ERR_PARAM;

	sf_twi_xfer_items[0].state       = 0;
	sf_twi_xfer_items[0].access_mode = BFIN_TWI_MODE_WRITE_REPEAT;
	sf_twi_xfer_items[0].count       = write_cnt;
	sf_twi_xfer_items[0].p_buffer    = (uint8_t*)p_write_data;
	sf_twi_xfer_items[1].state       = 0;
	sf_twi_xfer_items[1].access_mode = BFIN_TWI_MODE_READ;
	sf_twi_xfer_items[1].count       = read_cnt;
	sf_twi_xfer_items[1].p_buffer    = p_read_data;

	return e_bfin_twi_master_xfer(slave_addr, sf_twi_xfer_items, 2);
}

int32_t e_bfin_twi_master_write(uint8_t slave_addr, int32_t rstart, const uint8_t* p_write_data, uint16_t write_cnt)
{
	if(write_cnt <= 0) return BFIN_TWI_ERR_PARAM;

	sf_twi_xfer_items[0].state       = 0;
	if(BFIN_TWI_RSTART == rstart)
		sf_twi_xfer_items[0].access_mode = BFIN_TWI_MODE_WRITE_REPEAT;
	else
		sf_twi_xfer_items[0].access_mode = BFIN_TWI_MODE_WRITE;
	sf_twi_xfer_items[0].count       = write_cnt;
	sf_twi_xfer_items[0].p_buffer    = (uint8_t*)p_write_data;

	return e_bfin_twi_master_xfer(slave_addr, sf_twi_xfer_items, 1);
}

int32_t e_bfin_twi_master_read(uint8_t slave_addr, int32_t rstart, uint8_t* p_read_data, uint16_t read_cnt)
{
	if(read_cnt <= 0) return BFIN_TWI_ERR_PARAM;

	sf_twi_xfer_items[0].state       = 0;
	if(BFIN_TWI_RSTART == rstart)
		sf_twi_xfer_items[0].access_mode = BFIN_TWI_MODE_READ_REPEAT;
	else
		sf_twi_xfer_items[0].access_mode = BFIN_TWI_MODE_READ;
	sf_twi_xfer_items[0].count       = read_cnt;
	sf_twi_xfer_items[0].p_buffer    = p_read_data;

	return e_bfin_twi_master_xfer(slave_addr, sf_twi_xfer_items, 1);
}

int32_t e_bfin_twi_master_xfer_state(void)
{
	return sf_twi_xfer_param.state;
}

EX_INTERRUPT_HANDLER(s_bfin_twi_isr)
{
	uint16_t l_twi_stat = *pTWI_INT_STAT;
	*pTWI_INT_STAT = l_twi_stat;
	ssync();
	
	if((l_twi_stat & MERR) != 0)
	{
		sf_twi_xfer_param.state = BFIN_TWI_STATE_ERROR;
	}
	else if(sf_twi_xfer_param.state != BFIN_TWI_STATE_ERROR)
	{
		if((l_twi_stat & MCOMP) != 0)
		{
			if(sf_twi_xfer_param.item_count > 0)
			{
				sf_twi_xfer_param.item_count--;
				sf_twi_xfer_param.p_item = &sf_twi_xfer_param.p_item[1];
				// Start next item
				s_bfin_twi_master_xfer_start(sf_twi_xfer_param.p_item);
			}
			else
			{
				// Transfer done
				sf_twi_xfer_param.state = BFIN_TWI_STATE_COMP;
			}
		}
		else
		{
			if((BFIN_TWI_MODE_READ == sf_twi_xfer_param.p_item[0].access_mode) ||
				(BFIN_TWI_MODE_READ_REPEAT == sf_twi_xfer_param.p_item[0].access_mode))
			{
				// Read from queue
				while((sf_twi_xfer_param.p_item[0].count > 0) && ((*pTWI_FIFO_STAT & RCVSTAT) != 0))
				{
					*sf_twi_xfer_param.p_item[0].p_buffer = *pTWI_RCV_DATA8;
					sf_twi_xfer_param.p_item[0].p_buffer++;
					sf_twi_xfer_param.p_item[0].count--;
					ssync();
				}
			}
			else
			{
				// Write to queue
				while((sf_twi_xfer_param.p_item[0].count > 0) && ((*pTWI_FIFO_STAT & XMTSTAT) != XMTSTAT))
				{
					*pTWI_XMT_DATA8 = *sf_twi_xfer_param.p_item[0].p_buffer;
					sf_twi_xfer_param.p_item[0].p_buffer++;
					sf_twi_xfer_param.p_item[0].count--;
					ssync();
				}
			}
		}
	}
}
